#import sys
#sys.path.append('/home/ofercoq/src/cd_solver_root/')
import cd_solver
import numpy as np
norm = np.linalg.norm
import scipy.sparse as sp
import copy
from sklearn.datasets import load_svmlight_file, load_iris
import matplotlib.pyplot as plt

# get data

do_toy_problem = 0
if do_toy_problem:
    # min c.dot(x) : x >= 0, Mx <= b
    d = 3
    n = 4
    M = sp.csc_matrix(np.array([[2,4,5,7], [1,1,2,2], [1,2,3,3]]), dtype='float')
    c = -np.array([7,9,18,17])
    b = np.array([41,17,24])

    prob_toy = cd_solver.Problem(N=M.shape[1], f=['linear'], Af=c, g=['ineq_const'],
                                 blocks=[0, M.shape[1]],
                                 h=['ineq_const'], Ah=-M, bh=-b,
                                 blocks_h=[0, M.shape[0]])

    cd_solver.coordinate_descent(prob_toy, algorithm='pure-cd', verbose=0.01,
                                 print_style='smoothed_gap', average=True,
                                 restart_period=10,
                                 fixed_restart_period='adaptive',
                                 show_restart_info=False, 
                                 max_iter=5000)
if 0:
    # sparse SVM
    # min_w C sum_i max(0, 1 - y_i(x_i w)) + ||w||_1
    dataset = 'a1a'
    (X, y) = load_svmlight_file(dataset + '.data')

    #data = load_iris()
    #y = data.target
    #y[y==0] = -1
    #y[y==2] = -1
    #X = sp.csc_matrix(data.data)

    X = sp.csc_matrix(X.multiply(1./np.sqrt(X.multiply(X).sum(axis=0))))
    Xy = (X.T.multiply(y)).T
    lamb = 1


    # define problem
    prob = cd_solver.Problem(N=Xy.shape[1], f=['linear'], Af=-Xy.sum(axis=0),
                             cf=[0.5], bf=[-Xy.shape[0]],
                             g=['abs']*Xy.shape[1], cg=[lamb]*Xy.shape[1],
                             h=['abs']*Xy.shape[0], Ah=-Xy, ch=[0.5]*Xy.shape[0],
                             bh=[-1]*Xy.shape[0])
    prob2 = copy.copy(prob)
    problem = cd_solver.Problem(N=Xy.shape[1], f=['linear'], Af=-Xy.sum(axis=0),
                                cf=[0.5], bf=[-Xy.shape[0]], cg=[lamb],
                                g=['abs'], blocks=[0, Xy.shape[1]],
                                h=['abs'], Ah=-Xy, ch=[0.5], bh=[-1]*Xy.shape[0],
                                blocks_h=[0, Xy.shape[0]])
    problem2 = copy.copy(problem)

    # run pdhg
    #print('pdhg on iris')
    cd_solver.coordinate_descent(problem, algorithm='pure-cd', average=False,
                                 verbose=0.1, print_style='smoothed_gap',
                                 max_iter=50000, tolerance=1e-9)

    # run rapdhg
    print('pdhg with averaging and restart on '+dataset)
    cd_solver.coordinate_descent(problem2, algorithm='pure-cd', average=True,
                                 restart_period=30,
                                 verbose=0.1, print_style='smoothed_gap',
                                 max_iter=50000, tolerance=1e-9,
                                 fixed_restart_period='adaptive',
                                 show_restart_info=True)

    prints = np.array(problem.printed_values)
    plt.semilogy(prints[:,1], np.max(prints[:, 4:7], axis=1))
    prints = np.array(problem2.printed_values)
    plt.semilogy(prints[:,1], np.max(prints[:, 4:7], axis=1), linestyle='--')
    plt.xlabel('iteration')
    plt.ylabel('max(smoothed gap, dual infeasibility)')
    plt.legend(['PDHG', 'RAPDHG'])
    plt.show()

    # run pure-cd with averaging and restart
    print('pure-cd with averaging and restart on '+dataset)
    cd_solver.coordinate_descent(prob2, algorithm='pure-cd', average=True, restart_period=10, verbose=0.1, print_style='smoothed_gap', max_iter=20000, tolerance=1e-9, fixed_restart_period='adaptive')

    # run pure-cd
    print('pure-cd on '+dataset)
    cd_solver.coordinate_descent(prob, algorithm='pure-cd', average=False, verbose=0.1, print_style='smoothed_gap', max_iter=20000, tolerance=1e-9)

    prints = np.array(prob.printed_values)
    plt.semilogy(prints[:,1], np.max(prints[:, 4:7], axis=1))
    prints = np.array(prob2.printed_values)
    plt.semilogy(prints[:,1], np.max(prints[:, 4:7], axis=1), linestyle='--')
    plt.xlabel('iteration')
    plt.ylabel('max(smoothed gap, dual infeasibility)')
    plt.legend(['Pure-cd', 'RA-Pure-cd'])
    plt.show()


# run other algorithms

if 0:
    # ridge regression on real-sim
    (X, y) = load_svmlight_file('real-sim')
    # X2 = sp.csc_matrix(X.multiply(1./np.sqrt(X.multiply(X).sum(axis=0))))
    a = 100.
    prob_ridge = cd_solver.Problem(N=X.shape[1], h=['square'], Ah=X, bh=y, ch=[0.5],
                                 blocks_h=[0, X.shape[0]], g=['square'], cg=[a/2],
                                 blocks=[0, X.shape[1]])

    prob_ridge2 = copy.copy(prob_ridge)
    prob_ridge3 = cd_solver.Problem(N=X.shape[1], f=['square'], Af=X, bf=y, cf=[0.5],
                                 blocks_f=[0, X.shape[0]], g=['square'], cg=[a/2],
                                 blocks=[0, X.shape[1]])

        
    cd_solver.coordinate_descent(prob_ridge2, algorithm='pure-cd', verbose=1e-5, tolerance=1e-6, print_style='smoothed_gap', step_size_factor=1/2., average=True,
                                 restart_period=10, fixed_restart_period='adaptive')

    cd_solver.coordinate_descent(prob_ridge, algorithm='pure-cd', verbose=1e-5, tolerance=1e-6, print_style='smoothed_gap', step_size_factor=1/2.)

    cd_solver.coordinate_descent(prob_ridge3, algorithm='smart-cd', verbose=1e-5, tolerance=1e-6, print_style='smoothed_gap', restart_period=np.ceil(2*np.exp(1/2.)/np.sqrt(a)*sp.linalg.norm(X))-1, max_iter=1000, fixed_restart_period=True)

    prints = np.array(prob_ridge.printed_values)
    plt.semilogy(prints[:,1], np.max(prints[:, 4:7], axis=1))
    prints = np.array(prob_ridge2.printed_values)
    plt.semilogy(prints[:,1], np.max(prints[:, 4:7], axis=1), linestyle='--')
    prints = np.array(prob_ridge3.printed_values)
    plt.semilogy(prints[:,1], np.max(prints[:, 4:7], axis=1), linestyle='-.')
    plt.xlabel('iteration')
    plt.ylabel('duality gap')
    plt.legend(['PDHG', 'RAPDHG', 'restarted FISTA'])
    plt.show()


if 1:
    # L1 ROF
    # min_x ||Dx||_{2,1} + a ||x - im||_1
    im = plt.imread('cameraman.png')
    # im = im[:30,:30]
    N = np.prod(im.shape)
    # plt.imshow(im, cmap='gray')
    Dx = sp.diags([-np.ones(im.shape[0]), np.ones(im.shape[0])], offsets=[0, 1])
    Dx = sp.csc_matrix(Dx)[:-1]  # the last line is not a discrete gradient
    Dy = sp.diags([-np.ones(im.shape[1]), np.ones(im.shape[1])], offsets=[0, 1])
    Dy = sp.csc_matrix(Dy)[:-1]
    Dx = sp.kron(Dx, sp.eye(im.shape[1]))
    Dy = sp.kron(sp.eye(im.shape[0]), Dy)
    twoDgradient = sp.vstack([Dx, Dy], format='csc')
    twoDgradient.eliminate_zeros()
    # reorder the matrix
    twoDgradient = sp.csc_matrix((twoDgradient.data, 2 * (twoDgradient.indices % N) + twoDgradient.indices // N, twoDgradient.indptr), (2*N, N))
    a = 1.9

    skip_pdhg = True
    if skip_pdhg == False:
        pb_l1_rof = cd_solver.Problem(N=N, g=["abs"], cg=[a], bg=im.ravel(),
                    blocks=[0, N],
                    h=["norm2"] * N, blocks_h=np.arange(0, 2*N + 1, 2),
                                  Ah=twoDgradient, x_init=im.ravel())

        cd_solver.coordinate_descent(pb_l1_rof, algorithm='pure-cd', verbose=100, tolerance=1e-6, print_style='smoothed_gap', max_iter=1e6, step_size_factor=1, max_time=1e4, average=False)

    pb_l1_rof2 = cd_solver.Problem(N=N, g=["abs"], cg=[a], bg=im.ravel(),
                    blocks=[0, N],
                    h=["norm2"] * N, blocks_h=np.arange(0, 2*N + 1, 2),
                                   Ah=twoDgradient, x_init=im.ravel())

    print('run rapdhg on cameraman l1_rof with adaptive restart')
    cd_solver.coordinate_descent(pb_l1_rof2, algorithm='pure-cd', verbose=100, tolerance=1e-6, print_style='smoothed_gap', average=True, step_size_factor=1, restart_period=1000, fixed_restart_period='adaptive', max_iter=1e6, max_time=1e4, min_change_in_x=0, show_restart_info=True)

    def norm21(z):
        z = z.reshape((-1,2))
        return np.sum(np.sqrt(np.sum(z**2, axis=1)))
    
    def norm2inf(z):
        z = z.reshape((-1,2))
        return np.max(np.sqrt(np.sum(z**2, axis=1)))
    
    def value(x):
        return norm21(twoDgradient * x) + a * norm(x - im.ravel(), 1)

    def st(x, a):
        return np.sign(x) * np.maximum(0, np.abs(x) - a)
    
    def dual_value(y, beta, x_center):
        if norm2inf(y) > 1.00001:
            return np.inf
        DTy = twoDgradient.T.dot(y)
        dval = im.ravel().dot(DTy)
        x_bar = st(x_center + DTy/beta, a/beta)
        dval += DTy.dot(x_bar) - a*norm(x_bar, 1) - beta/2*norm(x_bar - x_center,2)**2
        return dval

    np.save('pb_l1_rof_solution', pb_l1_rof2.sol.reshape(im.shape))

    if skip_pdhg == True:
        prints = np.load('prints_pb_l1_rof.npy')
    else:
        prints = np.array(pb_l1_rof.printed_values)
        np.save('prints_pb_l1_rof', prints)
    plt.semilogy(prints[:,0], np.max(prints[:, 4:7], axis=1))

    prints = np.array(pb_l1_rof2.printed_values)
    np.save('prints_pb_l1_rof2', prints)
    plt.semilogy(prints[:,0], np.max(prints[:, 4:7], axis=1), linestyle='--')
    plt.xlabel('time (s)')
    plt.ylabel('max(smoothed gap, dual infeasibility)')
    plt.legend(['PDHG', 'RAPDHG'])
    plt.show()

    plt.imshow(pb_l1_rof2.sol.reshape(im.shape), cmap='gray'); plt.show()
    
"""
# test moyennes

sols = []
dual_sols = []
for k in range(5):
    cd_solver.coordinate_descent(problem, algorithm='pure-cd', average=False,
                             verbose=10, print_style='smoothed_gap',
                             max_iter=k+1, tolerance=1e-6)

    sols.append(problem.sol)
    dual_sols.append(problem.dual_sol)

cd_solver.coordinate_descent(problem, algorithm='pure-cd', average=True,
                             restart_period=5,
                             verbose=10, print_style='smoothed_gap',
                             max_iter=5, tolerance=1e-6)
"""
